class Heap:
    """Base class for heap object."""

    def __init__(self):
        """Initialize constructor for heap object."""
        self.heap = []

    def parent(self, index: int) -> int:
        return (index - 1) // 2

    def left_child(self, index: int):
        return (2 * index) + 1

    def right_child(self, index: int):
        return (2 * index) + 2

    def build_push_down(self, index: int):
        if min(self.heap):
            return self._build_push_down_min(index)
        else:
            return self._build_push_down_max(index)

    def build_floyd_heap(self, index: int):
        """Build Min-Heap based on Floyd's linear-time heap construction algorithm."""
        for index in enumerate(self.heap // 2):
            return build_push_down(self.heap, index)
        return self.heap

    def _build_push_down_iterative(self, index: int):
        # refactoring this code
        minimum = min(index)
        
        if minimum:
            if self.heap[minimum] > self.heap[index]:
                self.heap[minimum], self.heap[index] = self.heap[index], self.heap[minimum]
                if self.heap[minimum] < self.heap[self.parent]:
                    self.heap[minimum], self.heap[self.parent] = self.heap[self.parent], self.heap[minimum]
                    return _build_push_down_max(self.index, minimum)
        elif self.heap[minimum] < self.heap[index]:
            self.heap[minimum], self.heap[index] = self.heap[index], self.heap[minimum]

    def _build_push_up_heapify(self, index: int) -> int:
        """Bubble up algorithm."""
        if self.parent(index) > 0:
            if self.heap[self.parent(index)][1] > self.heap[index][1]:
                self.heap[self.parent(index)], self.heap[index] = self.heap[index], self.heap[self.parent(index)]
                self._build_push_up_heapify(self.parent(index))
                return self.heap
        return self.heap

    def _build_push_down_min(self, index: int):
        # this should be min(self.heap)
        # but it got an error
        # i dont know why this expression
        # still getting index out of range
        if min(self.heap[self.left_child(index)][1], self.heap[self.right_child(index)][1]) >= self.heap[index][1]:
            return self.heap

        if self.heap[self.left_child(index)][1] < self.heap[self.right_child(index)][1]:
            self.heap[self.left_child(index)], self.heap[index] = self.heap[index], self.heap[self.left_child(index)]
            self._build_push_down_min(self.left_child(index))
            return self.heap

        self.heap[self.right_child(index)], self.heap[index] = self.heap[index], self.heap[self.right_child(index)]
        self._build_push_down_min(self.right_child(index))
        return self.heap

    def _build_push_down_max(self, index: int):
        # refactoring this code
        maximum = max(index)

        if self.heap[maximum] > self.heap[index]:
            self.heap[minimum], self.heap[index] = self.heap[index], self.heap[maximum]

        if self.heap[maximum] < self.heap[self.parent]:
            self.heap[maximum], self.heap[self.parent] = self.heap[self.parent], self.heap[maximum]    

    def _build_push_up_min(self, index: int):
        if index[self.parent] and self.heap[index] < self.heap[self.parent]:
            self.heap[index], self.heap[self.parent] = self.heap[self.parent], self.heap[index]
            return _build_push_up_min(self.heap, self.parent(index))

    def _build_push_up_max(self, index: int):
        if index[self.parent] and self.heap[index] > self.heap[self.parent]:
            self.heap[index], self.heap[self.parent] = self.heap[self.parent], self.heap[index]
            return _build_push_up_max(self.heap, self.parent(index))

    def _build_push_up_iterative(self, index: int):
        while index[self.parent] and self.heap[index] < self.heap[self.parent]:
            self.heap[index], self.heap[self.parent] = self.heap[self.parent], self.heap[index]
            index = self.parent(index)

    def add(self, key: int, value: int) -> int:
        """Add and append the element in index."""
        self.heap.append((key, value))
        self._build_push_up_heapify(len(self.heap) - 1)

    def remove(self):
        """Remove minimum element in index."""
        minimum = self.heap[0]
        if minimum:
            minimum = self.heap.pop()
            self.build_push_down(0)
            return minimum

    def update(self, key: int, value: int):
        """Update key and value element in index."""
        for index, element in enumerate(self.heap):
            if element[0] == key:
                self.heap[index] = (key, value)
                if value >= element[1]:
                    self._build_push_down_min(index)
                else:
                    self._build_push_up_heapify(index)
                return self.heap
        return self.heap

    def remove_key(self, key):
        """Remove element in index based on their key."""
        for index, element in enumerate(self.heap):
            if element[0] == key:
                last_element = self.heap.pop()
                self.heap[index] = last_element
                # should return build_push_down
                # or returned as heapify ?
                self.build_push_down(index)
                return self.heap
        return self.heap


import time
from lru.heap import Heap
from abc import ABC, abstractmethod, ABCMeta

class BoundedLRUCache:

    __metaclass__ = ABCMeta

    @abstractmethod
    def clear_all(self):
        """Remove all element in cache dict."""

class LRUCache(BoundedLRUCache):

    def __init__(self, capacity: int=128):
        """Constructor for LRU Cache objects."""
        self.capacity = capacity
        self.dict = {}
        self.cache = Heap()

        if capacity is None:
            raise ValueError("Capacity must be set.")

    def __call__(self, *args, **kwargs):
        return self.cache

    def __len__(self):
        return len(self.capacity)

    def clear_all(self):
        return self.dict.clear()

    def clear_cache_key(self, key: int) -> int:
        """Clear cache in element based on their key."""
        return self.cache.remove_key(key)

    def get_cache(self, key: int):
        """Get cache in element based on their key, return
        `True` if the element has a key, otherwise return `False`
        when element hasn't a key.
        """
        if self.dict.get(key):
            return True
        return False

    def get_capacity(self):
        """Get cache capacity, return `True` if the cache
        is full otherwiser return `False` when the cache
        is not full.
        """
        if len(self.dict) > self.capacity:
            return True
        return False

    def set(self, key: int, value: int) -> int:
        """Set an objects that wants to be cached.
        
        :key: given key parameter as an integer

        :value: given value parameter of that key as an integer
        """
        end = time.perf_counter()
        start = time.perf_counter()
        access_time = start - end

        if self.get_cache(key):
            # update method is not working
            # should write update() method
            # in heap module
            self.cache.update(key, access_time)
            self.dict[key] = (value, access_time)
            return self.dict

        if self.get_capacity():
            minimum = self.cache.remove()[0]
            del self.dict[minimum]

        self.cache.add(key, access_time)
        self.dict[key] = (value, access_time)
        return self.dict

    def get(self, key: int):
        """Get the objects based on their key in cache element

        :key: given key parameter as an integer
        """
        if not self.get_cache(key):
            raise KeyError("Cache key not in elements.")

        access_time = time.perf_counter()
        self.cache.update(key, access_time)
        value = self.dict[key][0]
        self.dict[key] = (value, access_time)
        return value

    def get_lru_element(self):
        """Returned a dict type based on their key in cache element."""
        key = self.cache.heap[0][1]
        return self.dict[key]

    def get_dict(self):
        """Returned a dict type in cache element."""
        return self.dict

# mock test
test = LRUCache()

test.set(1, "test")
test.set(2, "foo")
test.set(3, "fc")
print(test.get_dict())
print(test.get_capacity())
print(test.get_cache(5))
# print(test.clear_cache(1))
print(test.clear_all())
print(test.get_dict())


import unittest
from nose.tools import raises
from lru.lrucache import LRUCache

class LRUCacheTest(unittest.TestCase):
    def setUp(self):
        self.capacity = 10
        self.cache = LRUCache(self.capacity)

        for i in range(self.capacity):
            self.cache.set(1, int(i))

    def tearDown(self):
        self.cache = None

    @raises(KeyError)
    def test_get(self):
        self.cache.set(1)


from lru.heap import Heap
from lru.lrucache import LRUCache

def lru_cache(capacity: int=128, **kwargs) -> int:
    """Decorators for LRUCache classes. Given the
    capacity of cache based on LRUCache classes, for example:
    
    ```
    @lru_cache(capacity=3)
    def foo(x):
        pass
    """
    def wrapper(func):
        return LRUCache(capacity=capacity, **kwargs)
    return wrapper


# mock test
"""
@lru_cache(capacity=5)
def test_lru(x):
    print("Calling f(" + str(x) + ")")
    return x

test_lru.set(1, "foo")
test_lru.set(2, "test")
test_lru.set(3, "foos")
test_lru.set(4, "fc")
test_lru.set(5, "set")
print(test_lru.get_capacity())
"""


import setuptools

import io
import os
import re

with io.open('lru/version.py', 'rt', encoding='utf-8') as f:
    version = re.search(r"__version__ = '(.*?)'", f.read()).group(1)

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="lruheap",
    version=version,
    author="Ryan Febriansyah",
    author_email="15523163@students.uii.ac.id",
    description="Python decorators for store in-data memory using LRU cache",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/sodrooome/lru-cache",
    license='MIT',
    zip_safe=False,
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=3.5',
)